import numpy as np
import pandas as pd

from montre.algebra import TimedRelation, Bound
#
# Timed behavior implementation based on Pandas' dataframes
# 

class TimedBehavior(object):
    """A container for timed behaviors based on Pandas' dataframes"""
    def __init__(self, dates, values, columns=None):
        super(TimedBehavior, self).__init__()

        self.df = pd.DataFrame(values, index=pd.to_datetime(dates), columns=columns, dtype="category")

    def merge(self, other):
        m = pd.concat([self.df, other.df], axis=1, ignore_index=True).fillna(method='bfill')
        return TimedBehavior(m.index, list(m.values), columns=list(self.df.columns)+list(other.df.columns))

    def __str__(self):
        return str(self.df)


    def to_timed_relation(self, column, category):
        """Naive builder"""
        
        Z = TimedRelation()
        begin, end = 0, 0
        # code = self.df[column].cat.categories.get_loc(category)
        # codes = self.df[column].cat.codes
        dates = self.df[column].index.view('int64')
        values = self.df[column].values
        active = False

        for i in np.arange(0, len(dates)):
            #print(i, dates[i], codes[i], values[i])
            if values[i] == category and active == False:
                active = True
                begin = dates[i-1]
            elif values[i] != category and active == True:
                active = False
                end = dates[i-1]
                Z.append(
                    bmin=Bound.geq(begin),
                    bmax=Bound.leq(end),
                    emin=Bound.geq(begin),
                    emax=Bound.leq(end),
                    dmin=Bound.gt(0)
                )

        if active:
            end = dates[-1]
            Z.append(
                    bmin=Bound.geq(begin),
                    bmax=Bound.leq(end),
                    emin=Bound.geq(begin),
                    emax=Bound.leq(end),
                    dmin=Bound.gt(0)
                )

        return Z




import os
import sys

from ctypes import cdll, c_int64, c_void_p, POINTER

from distutils.sysconfig import get_python_lib

class TimedRelation:

    if os.name == 'nt':
        libmontre = windll.LoadLibrary(os.path.join(get_python_lib(), "libmontre")) # NOT TESTED YET
    else:
        try:
            libmontre = cdll.LoadLibrary(os.path.join("..", "libmontre.so")) # DEVELOPMENT
            print("Hello")
        except OSError:
            libmontre = cdll.LoadLibrary(os.path.join(get_python_lib(), "libmontre.so")) # RELEASE
        except:
            print("libmontre.so couldn't found.") 

    zone_type = POINTER(c_int64)

    libmontre.zs_create.restype = c_void_p
    libmontre.zs_create.argtypes = []
    libmontre.zs_destroy.restype = None
    libmontre.zs_destroy.argtypes = []
    libmontre.zs_size.restype = c_int64
    libmontre.zs_includes.restype = c_int64

    libmontre.zs_append.argtypes = [c_void_p, c_int64, c_int64, c_int64, c_int64,c_int64,c_int64]
    libmontre.zs_append.restype = None

    libmontre.zs_append_not_anchored.argtypes = [c_void_p, c_int64, c_int64]
    libmontre.zs_append_not_anchored.restype = None

    libmontre.zs_get_zone.restype = zone_type
    libmontre.zs_get_zone.argtypes = [c_void_p, c_int64]

    def __init__(self, obj=None):
        if obj == None:
            self.obj = TimedRelation.libmontre.zs_create()
        else:
            self.obj = obj

    def __del__(self):
        TimedRelation.libmontre.zs_destroy(self.obj)

    def __len__(self):
        return TimedRelation.libmontre.zs_size(self.obj)

    def __iter__(self):
        def get_zone(self, i):
            z = TimedRelation.libmontre.zs_get_zone(self.obj, i)
            return (z[1], z[3], z[2], z[6], z[5], z[7])
        return (get_zone(self, i) for i in range(len(self)))

    @property
    def zones(self):
        return [((-(z[0]//2), z[1]//2, -(z[2]//2), z[3]//2, -(z[4]//2), z[5]//2), (z[0]%2, z[1]%2, z[2]%2, z[3]%2, z[4]%2, z[5]%2)) for z in self]

    def __str__(self):
        return "\n".join([
            "({bminval}{bminbnd}x{bmaxbnd}{bmaxval}, {eminval}{eminbnd}y{emaxbnd}{emaxval}, {dminval}{dminbnd}y-x{dmaxbnd}{dmaxval})".format(
            bminval=v[0], bminbnd="<" if b[0] == 0 else "<=",
            bmaxval=v[1], bmaxbnd="<" if b[1] == 0 else "<=",
            eminval=v[2], eminbnd="<" if b[2] == 0 else "<=",
            emaxval=v[3], emaxbnd="<" if b[3] == 0 else "<=",
            dminval=v[4], dminbnd="<" if b[4] == 0 else "<=",
            dmaxval=v[5], dmaxbnd="<" if b[5] == 0 else "<=",
            ) for v, b in self.zones]
        )
        # return str([values for values, bounds in self.zones])

    # def __getitem__(self, i):  # access elements in vector at index
    #     if 0 <= i < len(self):
    #         return libmontre.zs_get_zone(self.obj, i)
    #     raise IndexError('Vector index out of range')

    def append(self, bmin=0, bmax=sys.maxsize, emin=0, emax=sys.maxsize, dmin=0, dmax=sys.maxsize):
        TimedRelation.libmontre.zs_append(self.obj, bmin, bmax, emin, emax, dmin, dmax)
        return self

    def append_not_anchored(self, begin, end):
        TimedRelation.libmontre.zs_append_not_anchored(self.obj, begin, end)
        return self

    def absorb(self):
        retobj = TimedRelation.libmontre.zs_create()
        TimedRelation.libmontre.zs_filter(retobj, self.obj)
        self.obj = retobj

    @staticmethod
    def absorbed(self):
        retobj = TimedRelation.libmontre.zs_create()
        TimedRelation.libmontre.zs_filter(retobj, self.obj)
        return TimedRelation(retobj)

    def includes(self, other):
        return TimedRelation.libmontre.zs_includes(self.obj, other.obj) != 0

    def restrict(self, a, b):
        retobj = TimedRelation.libmontre.zs_create()
        TimedRelation.libmontre.zs_restrict(retobj, self.obj, c_int64(a), c_int64(b))
        return TimedRelation(retobj)

    def intersect(self, other):
        retobj = TimedRelation.libmontre.zs_create()
        TimedRelation.libmontre.zs_intersect(retobj, self.obj, other.obj)
        return TimedRelation(retobj)

    def concatenate(self, other):
        retobj = TimedRelation.libmontre.zs_create()
        TimedRelation.libmontre.zs_concatenate(retobj, self.obj, other.obj)
        return TimedRelation(retobj)

    def union(self, other):
        retobj = TimedRelation.libmontre.zs_create()
        TimedRelation.libmontre.zs_union(retobj, self.obj, other.obj)
        return TimedRelation(retobj)

    def plus(self):
        retobj = TimedRelation.libmontre.zs_create()
        TimedRelation.libmontre.zs_plus(retobj, self.obj)
        return TimedRelation(retobj)


    def __xor__(self, other):
        return self.concatenate(other)

    def __mod__(self, time_bounds):
        return self.restrict(time_bounds[0], time_bounds[1])

    def __or__(self, other):
        return self.union(other)

    def __and__(self, other):
        return self.intersect(other)

    def __mul__(self, other=None):
        return self.plus()

class Bound:
    @staticmethod
    def lt(c):
        return 2*c
    @staticmethod
    def leq(c):
        return 2*c + 1
    @staticmethod
    def gt(c):
        return -2*c
    @staticmethod
    def geq(c):
        return -2*c + 1




import os
import sys

from os.path import join

from distutils.core import setup, Extension
from distutils.command.build_ext import build_ext

class build_ext(build_ext):

    def build_extension(self, ext):
        self._ctypes = isinstance(ext, CTypes)
        return super().build_extension(ext)

    def get_export_symbols(self, ext):
        if self._ctypes:
            return ext.export_symbols
        return super().get_export_symbols(ext)

    def get_ext_filename(self, ext_name):
        if self._ctypes:
            return ext_name + '.so'
        return super().get_ext_filename(ext_name)

class CTypes(Extension): pass

module1 = CTypes('libmontre',
            define_macros = [ ('MAJOR_VERSION', '0'),
                              ('MINOR_VERSION', '6')],
            sources = [ join("src", "libmontre", "zone2.cpp"),
                        join("src", "libmontre", "zoneset.cpp"),
                        join("src", "libmontre", "boolean.cpp"),
                        join("src", "libmontre", "interval.cpp")],
            depends = [ join("src", "libmontre", "zone2.h"),
                        join("src", "libmontre", "zoneset.h"),
                        join("src", "libmontre", "boolean.h"),
                        join("src", "libmontre", "interval.h")],
            include_dirs=[join("src", "libmontre")],
            extra_compile_args=['-std=c++11'])

setup (name = 'montre',
       version='0.0.2',
       author='Dogan Ulus',
       author_email='doganulus@gmail.com',
       packages=["montre"],
       license='GPLv3+',
       classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3',
    ],
       description = 'A timed pattern matching package',
       url='http://pypi.python.org/pypi/montre/',
       py_modules = ["montre.algebra", "montre.behavior"],
       ext_modules = [module1],
       data_files= [join("src", "libmontre", "zone2.h"),
                    join("src", "libmontre", "zoneset.h"),
                    join("src", "libmontre", "boolean.h"),
                    join("src", "libmontre", "interval.h")],
       cmdclass={'build_ext': build_ext})

